from ecommercetools.customers.customers import get_customers
from ecommercetools.customers.customers import get_rfm_segments
from ecommercetools.customers.customers import get_abc_segments
from ecommercetools.customers.customers import get_cohorts
from ecommercetools.customers.customers import get_cohort_matrix
from ecommercetools.customers.customers import get_retention
from ecommercetools.customers.customers import get_latency
from ecommercetools.customers.customers import get_customer_predictions



import pandas as pd
from ecommercetools.utilities import tools


def get_products(transaction_items, days=None):
    """Return a Pandas DataFrame of products from a Pandas DataFrame of transaction items.

    Args:
        transaction_items (object): Pandas DataFrame.
        days (int, optional): Select only product sold in the last X days.

    Returns:
        customers (object): Pandas DataFrame
    """

    if days:
        transaction_items = tools.select_last_x_days(transaction_items, 'order_date', days)

    transaction_items = transaction_items.assign(line_price=transaction_items['quantity'] * transaction_items['unit_price'])

    products = transaction_items.groupby('sku').agg(
        first_order_date=('order_date', 'min'),
        last_order_date=('order_date', 'max'),
        customers=('customer_id', 'nunique'),
        orders=('order_id', 'nunique'),
        items=('quantity', 'sum'),
        revenue=('line_price', 'sum'),
        avg_unit_price=('unit_price', 'mean'),
        avg_quantity=('quantity', 'mean'),
        avg_revenue=('line_price', 'mean')
    ).reset_index()

    products['avg_orders'] = round(products['orders'] / products['customers'], 2)
    products['product_tenure'] = (pd.to_datetime('today') - products['first_order_date']).dt.days
    products['product_recency'] = (pd.to_datetime('today') - products['last_order_date']).dt.days
    return products


def get_repurchase_rate_label(df):
    """Add a label describing the repurchase rate bin.

    Args:
        df (object): Pandas DataFrame containing repurchase_rate.

    Returns:
    -------
        df (object): Pandas DataFrame with repurchase_rate_label added.
    """

    labels = ['Very low repurchase',
              'Low repurchase',
              'Moderate repurchase',
              'High repurchase',
              'Very high repurchase']
    df['repurchase_rate_label'] = pd.cut(df['repurchase_rate'],
                                         bins=5,
                                         labels=labels)
    return df


def get_bulk_purchase_rate_label(df):
    """Add a label describing the bulk purchase rate bin.

    Args:
        df (object): Pandas DataFrame containing bulk_purchase_rate.

    Returns:
    -------
        df (object): Pandas DataFrame with bulk_purchase_rate_label added.
    """

    labels = ['Very low bulk',
              'Low bulk',
              'Moderate bulk',
              'High bulk',
              'Very high bulk']
    df['bulk_purchase_rate_label'] = pd.cut(df['bulk_purchase_rate'],
                                            bins=5,
                                            labels=labels)
    return df


def get_repurchase_rates(df):
    """Return repurchase rates and purchase behaviour for each SKU from transaction items data.

    Given a Pandas DataFrame of transactional items, this function returns a Pandas DataFrame
    containing the purchase behaviour and repurchase behaviour for each SKU.

    Args:
        df (object): Pandas DataFrame. Required columns: sku, order_id, customer_id, quantity, unit_price.

    Returns:
    -------
        df (object): Pandas DataFrame.
    """

    # Count the number of times each customer purchased each SKU
    df['times_purchased'] = df.groupby(['sku', 'customer_id'])['order_id'].transform('count')

    # Count the number of times the SKU was purchased individually within orders
    df['purchased_individually'] = df[df['quantity'] == 1]. \
        groupby('sku')['order_id'].transform('count')
    df['purchased_individually'] = df['purchased_individually'].fillna(0)

    # Count the number of times the SKU was purchased once only by customers
    df['purchased_once'] = df[df['times_purchased'] == 1]. \
        groupby('sku')['order_id'].transform('count')
    df['purchased_once'] = df['purchased_once'].fillna(0)

    # Calculate line price
    df['line_price'] = df['unit_price'] * df['quantity']

    # Get unique SKUs and count total items, orders, and customers
    df_skus = df.groupby('sku').agg(
        revenue=('line_price', 'sum'),
        items=('quantity', 'sum'),
        orders=('order_id', 'nunique'),
        customers=('customer_id', 'nunique'),
        avg_unit_price=('unit_price', 'mean'),
        avg_line_price=('line_price', 'mean')
    )

    # Calculate the average number of units per order
    df_skus = df_skus.assign(avg_items_per_order=(df_skus['items'] / df_skus['orders']))

    # Calculate the average number of items per customer
    df_skus = df_skus.assign(avg_items_per_customer=(df_skus['items'] / df_skus['customers']))

    # Merge the dataframes
    df_subset = df[['sku', 'purchased_individually', 'purchased_once']].fillna(0)
    df_subset.drop_duplicates('sku', keep='first', inplace=True)
    df_skus = df_skus.merge(df_subset, on='sku', how='left')

    # Calculate bulk purchase rates
    df_skus = df_skus.assign(bulk_purchases=(df_skus['orders'] - df_skus['purchased_individually']))
    df_skus = df_skus.assign(bulk_purchase_rate=(df_skus['bulk_purchases'] / df_skus['orders']))

    # Calculate repurchase rates
    df_skus = df_skus.assign(repurchases=(df_skus['orders'] - df_skus['purchased_once']))
    df_skus = df_skus.assign(repurchase_rate=(df_skus['repurchases'] / df_skus['orders']))

    # Add labels
    df_skus = get_repurchase_rate_label(df_skus)
    df_skus = get_bulk_purchase_rate_label(df_skus)

    df_skus['bulk_and_repurchase_label'] = df_skus['repurchase_rate_label'].astype(str) + \
                                           '_' + df_skus['bulk_purchase_rate_label'].astype(str)

    return df_skus



"""
Fetch the contents of all XML sitemaps and return the output in a Pandas dataframe.
"""

import pandas as pd
import urllib.request
from urllib.parse import urlparse
from bs4 import BeautifulSoup


def _get_xml(url: str):
    """Scrapes an XML sitemap from the provided URL and returns XML source.

    Args:
        url (string): Fully qualified URL pointing to XML sitemap.

    Returns:
        xml (string): XML source of scraped sitemap.
    """

    try:
        response = urllib.request.urlopen(url)
        xml = BeautifulSoup(response,
                            'lxml-xml',
                            from_encoding=response.info().get_param('charset'))
        return xml
    except Exception as e:
        print("Error: ", e)


def _get_sitemap_type(xml: str):
    """Parse XML source and returns the type of sitemap.

    Args:
        xml (string): Source code of XML sitemap.

    Returns:
        sitemap_type (string): Type of sitemap (sitemap, sitemapindex, or None).
    """

    sitemapindex = xml.find_all('sitemapindex')
    sitemap = xml.find_all('urlset')

    if sitemapindex:
        return 'sitemapindex'
    elif sitemap:
        return 'urlset'
    else:
        return


def _get_child_sitemaps(xml: str):
    """Return a list of child sitemaps present in a XML sitemap file.

    Args:
        xml (string): XML source of sitemap.

    Returns:
        sitemaps (list): Python list of XML sitemap URLs.
    """

    sitemaps = xml.find_all("sitemap")
    output = []

    for sitemap in sitemaps:
        output.append(sitemap.findNext("loc").text)
    return output


def _sitemap_to_dataframe(xml: str, name=None, verbose=False):
    """Read an XML sitemap into a Pandas dataframe.

    Args:
        xml (string): XML source of sitemap.
        name (optional): Optional name for sitemap parsed.
        verbose (boolean, optional): Set to True to monitor progress.

    Returns:
        dataframe: Pandas dataframe of XML sitemap content.
    """

    df = pd.DataFrame(columns=['loc', 'changefreq', 'priority', 'domain', 'sitemap_name'])

    urls = xml.find_all("url")

    for url in urls:

        if xml.find("loc"):
            loc = url.findNext("loc").text
            parsed_uri = urlparse(loc)
            domain = '{uri.netloc}'.format(uri=parsed_uri)
        else:
            loc = ''
            domain = ''

        if xml.find("changefreq"):
            changefreq = url.findNext("changefreq").text
        else:
            changefreq = ''

        if xml.find("priority"):
            priority = url.findNext("priority").text
        else:
            priority = ''

        if name:
            sitemap_name = name
        else:
            sitemap_name = ''

        row = {
            'domain': domain,
            'loc': loc,
            'changefreq': changefreq,
            'priority': priority,
            'sitemap_name': sitemap_name,
        }

        if verbose:
            print(row)

        df = df.append(row, ignore_index=True)
    return df


def get_sitemap(url: str):
    """Return a dataframe containing all of the URLs from a site's XML sitemaps.

    Args:
        url (string): URL of site's XML sitemap. Usually located at /sitemap.xml

    Returns:
        df (dataframe): Pandas dataframe containing all sitemap content.

    """

    xml = _get_xml(url)
    if xml:
        sitemap_type = _get_sitemap_type(xml)

        if sitemap_type =='sitemapindex':
            sitemaps = _get_child_sitemaps(xml)
        else:
            sitemaps = [url]

        df = pd.DataFrame(columns=['loc', 'changefreq', 'priority', 'domain', 'sitemap_name'])

        for sitemap in sitemaps:
            sitemap_xml = _get_xml(sitemap)
            df_sitemap = _sitemap_to_dataframe(sitemap_xml, name=sitemap)

            df = pd.concat([df, df_sitemap], ignore_index=True)

        return df




"""
Fetch the contents of a robots.txt file and return the output in a Pandas dataframe.
"""

import requests
import urllib.parse
import json
import pandas as pd
from requests_html import HTMLSession


def _get_source(url: str):
    """Return the source code for the provided URL.

    Args:
        url (string): URL of the page to scrape.

    Returns:
        response (object): HTTP response object from requests_html.
    """

    try:
        session = HTMLSession()
        response = session.get(url)
        return response
    except requests.exceptions.RequestException as e:
        print(e)


def get_sitemaps(url: str):
    """Parse a robots.txt file and return a Python list containing any sitemap URLs found.

    Args:
        url (string): URL of robots.txt file.

    Returns:
        data (list): List containing each sitemap found.
    """

    response = _get_source(url)
    robots = response.text

    data = []
    lines = str(robots).splitlines()

    for line in lines:
        if line.startswith('Sitemap:'):
            split = line.split(':', maxsplit=1)
            data.append(split[1].strip())

    return data


def get_robots(url: str):
    """Parses robots.txt file contents into a Pandas DataFrame.

    Args:
        url (string): URL of robots.txt file.

    Returns:
        df (list): Pandas dataframe containing robots.txt directives and parameters.
    """

    response = _get_source(url)
    robots = response.text

    data = []
    lines = str(robots).splitlines()
    for line in lines:

        if line.strip():
            if not line.startswith('#'):
                split = line.split(':', maxsplit=1)
                data.append([split[0].strip(), split[1].strip()])

    return pd.DataFrame(data, columns=['directive', 'parameter'])




import pandas as pd
from ecommercetools.products import products


def _abc_classify_product(percentage):
    """Return an ABC classification for a product based on its ranked percentage revenue contribution.

    Args:
        percentage (float): Running percentage of revenue contributed by each SKU over a time period.

    Returns:
        class (string): ABC class string
    """

    if 0 < percentage <= 80:
        return 'A'
    elif 80 < percentage <= 90:
        return 'B'
    else:
        return 'C'


def get_inventory_classification(transaction_items, days=None, verbose=False):
    """Return a Pandas DataFrame of product inventory classification from the transaction items dataframe.

    Args:
        transaction_items (object): Pandas DataFrame of transaction items.
        days (int, optional): Return data only for products sold in the past X days.
        verbose (bool, optional): Displays additional columns of workings when set to True.

    Returns:
        products (object): Pandas DataFrame.
    """

    # Filter to the last X days
    if days:
        products_data = products.get_products(transaction_items, days)
    else:
        products_data = products.get_products(transaction_items)

    # Sort the data
    products_data['revenue_total'] = products_data['revenue'].sum()
    products_data = products_data.sort_values(by='revenue', ascending=False)

    # ABC inventory classification
    products_data['revenue_cumsum'] = products_data['revenue'].cumsum()
    products_data['revenue_running_percentage'] = (products_data['revenue_cumsum'] / products_data['revenue_total']) * 100
    products_data['abc_class'] = products_data['revenue_running_percentage'].apply(_abc_classify_product)
    products_data['abc_rank'] = products_data['revenue_running_percentage'].rank().astype(int)

    if verbose:
        products_data = products_data[['sku', 'abc_class', 'abc_rank', 'revenue',
                                       'revenue_cumsum', 'revenue_total', 'revenue_running_percentage']]
    else:
        products_data = products_data[['sku', 'abc_class', 'abc_rank']]

    return products_data



"""
Fetch data from the Google Search Console API.
"""

import sys
import pandas as pd
from google.oauth2 import service_account
from googleapiclient.discovery import build


def _connect(key: str):
    """Create a connection to the Google Search Console API and return service object.

    Args:
        key (string): Google Search Console JSON client secrets path.

    Returns:
        service (object): Google Search Console service object.
    """

    try:
        scope = ['https://www.googleapis.com/auth/webmasters']
        credentials = service_account.Credentials.from_service_account_file(key, scopes=scope)
        service = build('webmasters', 'v3', credentials=credentials)

        return service

    except Exception as e:
        print("Error: ", e)
        sys.exit(1)


def query_google_search_console(key: str, site_url: str, payload: dict):
    """Run a query on the Google Search Console API and return a dataframe of results.

    Args:
        key (object): JSON client secrets key file path.
        site_url (string): URL of Google Search Console property
        payload (dict): API query payload dictionary

    Return:
        df (dataframe): Pandas dataframe containing requested data.
    """

    service = _connect(key)

    response = service.searchanalytics().query(siteUrl=site_url, body=payload).execute()

    results = []

    for row in response['rows']:
        data = {}

        for i in range(len(payload['dimensions'])):
            data[payload['dimensions'][i]] = row['keys'][i]

        data['clicks'] = row['clicks']
        data['impressions'] = row['impressions']
        data['ctr'] = round(row['ctr'] * 100, 2)
        data['position'] = round(row['position'], 2)
        results.append(data)

    return pd.DataFrame.from_dict(results)





import requests
import urllib.parse
import json
import pandas as pd
from requests_html import HTMLSession


def _get_source(url: str):
    """Return the source code for the provided URL.

    Args:
        url (string): URL of the page to scrape.

    Returns:
        response (object): HTTP response object from requests_html.
    """

    try:
        session = HTMLSession()
        response = session.get(url)
        return response

    except requests.exceptions.RequestException as e:
        print(e)


def get_knowledge_graph(api_key: str,
                        query: str,
                        output="dataframe"):
    """Return a Google Knowledge Graph for a given query.

    Args:
        api_key (string): Google Knowledge Graph API key.
        query (string): Term to search for.
        output (string, optional): Output format (dataframe, or json).

    Returns:
        response (object): Knowledge Graph response object in JSON format.
    """

    endpoint = 'https://kgsearch.googleapis.com/v1/entities:search'
    params = {
        'query': query,
        'limit': 10,
        'indent': True,
        'key': api_key,
    }

    url = endpoint + '?' + urllib.parse.urlencode(params)
    response = _get_source(url)

    if output == "json":
        return json.loads(response.text)
    else:
        return pd.json_normalize(json.loads(response.text), record_path='itemListElement')



"""
Get keyword suggestions for a term using Google Autocomplete or Google Suggest.
"""

import requests
import urllib.parse
import json
import pandas as pd
from requests_html import HTMLSession


def _get_source(url: str):
    """Return the source code for the provided URL.

    Args:
        url (string): URL of the page to scrape.

    Returns:
        response (object): HTTP response object from requests_html.
    """

    try:
        session = HTMLSession()
        response = session.get(url)
        return response
    except requests.exceptions.RequestException as e:
        print(e)


def _get_results(query: str):
    """Get the JSON data from a Google Autocomplete query.

    Args:
        query (string): Query term, i.e. data science

    Returns:
        results (dict): JSON results.
    """

    query = urllib.parse.quote_plus(query)
    response = _get_source("https://suggestqueries.google.com/complete/search?output=chrome&hl=en&q=" + query)
    results = json.loads(response.text)
    return results


def _format_results(results: dict):
    """Return formatted dictionary containing term and relevance.

    Args:
        results (dict): JSON dictionary of Google Autocomplete results.

    Returns:
        suggestions (dict): Formatted dictionary containing term and relevance.
    """

    if results:
        suggestions = []
        for index, value in enumerate(results[1]):
            suggestion = {'term': value, 'relevance': results[4]['google:suggestrelevance'][index]}
            suggestions.append(suggestion)
        return suggestions


def _get_suggestions(query: str):
    """Return results sorted by relevance.

    Args:
        query (string): Search term, i.e. data science

    Returns:
        results (dict): Sorted dictionary containing term and relevance.
    """

    results = _get_results(query)
    results = _format_results(results)
    results = sorted(results, key=lambda k: k['relevance'], reverse=True)
    return results


def _get_expanded_term_suffixes():
    """Return a list of query suffixes to extend Google Autocomplete results.

    Returns:
        expanded_term_suffixes (list)
    """

    expanded_term_suffixes = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
                              'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']
    return expanded_term_suffixes


def _get_expanded_term_prefixes():
    """Return a list of query prefixes to extend Google Autocomplete results.

    Returns:
        expanded_term_prefixes (list)
    """

    expanded_term_prefixes = ['who is *', 'what is *', 'where is *', 'when can *', 'why is *',
                              'how to *', 'best', 'cheap', 'worst', 'is', 'what', 'when', 'why',
                              'how', 'who']
    return expanded_term_prefixes


def _get_expanded_terms(query: str):
    """Return a list of expanded terms, comprising the original query, and the prefixed and suffixed queries.

    Args:
        query (string): Query term, i.e. data science

    Returns:
        terms (list): List of query terms with suffixes and prefixes.
    """

    expanded_term_prefixes = _get_expanded_term_prefixes()
    expanded_term_suffixes = _get_expanded_term_suffixes()

    terms = [query]

    for term in expanded_term_prefixes:
        terms.append(term + ' ' + query)

    for term in expanded_term_suffixes:
        terms.append(query + ' ' + term)

    return terms


def _get_expanded_suggestions(query: str):
    """Return the Google Autocomplete suggestions for a query and its prefixed and suffixed versions.

    Args:
        query (string): Query term, i.e. data science

    Returns:
        all_results (dict): Sorted formatted dictionary of results for each search term.
    """

    all_results = []

    expanded_terms = _get_expanded_terms(query)
    for term in expanded_terms:
        results = _get_results(term)
        results = _format_results(results)
        all_results = all_results + results
        all_results = sorted(all_results, key=lambda k: k['relevance'], reverse=True)
    return all_results


def google_autocomplete(query: str, include_expanded=True):
    """Run a Google Autocomplete / Google Suggest search with optional query expansion.

    Args:
        query (string): Query term, i.e. data science
        include_expanded (bool, optional): Optional boolean flag. Set to true to add prefixes/suffixes.

    Returns:
        df (dataframe): Pandas dataframe containing results.
    """

    if include_expanded:
        results = _get_expanded_suggestions(query)

    else:
        results = _get_suggestions(query)

    df = pd.DataFrame.from_records(results)
    return df




import pandas as pd
import numpy as np
from datetime import timedelta, datetime


def load_transaction_items(filename,
                           date_column='order_date',
                           order_id_column='order_id',
                           customer_id_column='customer_id',
                           sku_column='sku',
                           quantity_column='quantity',
                           unit_price_column='unit_price'
                           ):
    """Load a CSV of transactional item data, sets standard column names, and calculates line price.

    Args:
        filename (str): Filename and path of CSV file containing transaction items.
        date_column (str, optional): Name of order date column, default is order_date
        order_id_column (str, optional): Name of order ID column, default is order_id
        customer_id_column (str, optional): Name of customer ID column, default is customer_id
        sku_column (str, optional): Name of SKU column, default is sku
        quantity_column (int, optional): Name of quantity column, default is quantity
        unit_price_column (float, optional): Name of unit price column, default is unit_price

    Usage:
        transaction_items = rt.load_data('data/input/transaction_items_non_standard_names.csv',
                                     date_column='InvoiceDate',
                                     order_id_column='OrderId',
                                     customer_id_column='CustomerId',
                                     sku_column='VariantId',
                                     quantity_column='Qty',
                                     unit_price_column='Price'
                                     )

    Returns:
        A Pandas dataframe containing the same data with the column names changed to the
        standardised names used throughout RetailTools, if they do not already match, and
        the order_date column correctly set as a datetime column. If the user provides a
        CSV file in which the column names are already set to these values, it it not a
        requirement to provide them.

    """

    df = pd.read_csv(filename, parse_dates=[date_column])
    df = df.rename(columns={
        date_column: 'order_date',
        order_id_column: 'order_id',
        customer_id_column: 'customer_id',
        sku_column: 'sku',
        quantity_column: 'quantity',
        unit_price_column: 'unit_price'
    })
    df['line_price'] = round(df['quantity'] * df['unit_price'], 2)
    return df


def load_sample_data():
    """Load the Online Retail dataset of transaction items and format for use within EcommerceTools functions.

    :return: Pandas dataframe.
    """

    df = pd.read_csv('https://raw.githubusercontent.com/databricks/Spark-The-Definitive-Guide/master/data'
                     '/retail-data/all/online-retail-dataset.csv',
                     names=['order_id', 'sku', 'description', 'quantity', 'order_date', 'unit_price', 'customer_id',
                            'country'],
                     skiprows=1,
                     parse_dates=['order_date']
                     )
    df['line_price'] = df['unit_price'] * df['quantity']
    return df


def get_cumulative_count(df, group_column, count_column, sort_column):
    """Get the cumulative count of a column based on a GroupBy.

    Args:
        df (object): Pandas DataFrame.
        group_column (string): Column to group by.
        count_column (string): Column to count.
        sort_column (string): Column to sort by.

    Returns:
        Cumulative count of the column.

    Usage:
        df['running_total'] = get_cumulative_count(df, 'customer_id', 'order_id', 'date_created')
    """

    df = df.sort_values(by=sort_column, ascending=True)
    return df.groupby([group_column])[count_column].cumcount()


def get_previous_value(df, group_column, value_column):
    """Group by a column and return the previous value of another column and assign value to a new column.

    Args:
        df (object): Pandas DataFrame.
        group_column (str): Column name to group by
        value_column (str): Column value to return.

    Returns:
        Original DataFrame with new column containing previous value of named column.
    """

    df = df.copy()
    df = df.sort_values(by=[value_column], ascending=False)
    return df.groupby([group_column])[value_column].shift(-1)


def get_days_since_date(df, before_datetime, after_datetime):
    """Return a new column containing the difference between two dates in days.

    Args:
        df (object): Pandas DataFrame.
        before_datetime (datetime): Earliest datetime (will convert value)
        after_datetime (datetime): Latest datetime (will convert value)

    Returns:
        New column value
    """

    df = df.copy()
    df[before_datetime] = pd.to_datetime(df[before_datetime])
    df[after_datetime] = pd.to_datetime(df[after_datetime])

    diff = df[after_datetime] - df[before_datetime]
    return round(diff / np.timedelta64(1, 'D')).fillna(0).astype(int)


def date_subtract(date, days):
    """Given a date, subtract a specified number of days, and return the date.

    Args:
        date (datetime): Original date to subtract from.
        days (int): Number of days to subtract from date.

    Return:
        subtracted_date (datetime): Original date with days subtracted.
    """

    return pd.to_datetime(date) - timedelta(days=days)


def select_last_x_days(df,
                       date_column='order_date',
                       days=365):
    """Select the last X days from a Pandas dataframe.

    Args:
        df (object): Pandas dataframe containing time series data.
        date_column (str, optional): Name of column containing date. Default is order_date.
        days (int, optional): Number of days to subtract from current date. Default is 365.

    Returns:
        df (object): Filtered dataframe containing only records from the past X days.
    """

    subtracted_date = date_subtract(datetime.today(), days)
    df = df[df[date_column] >= subtracted_date]
    return df



from ecommercetools.seo.robots import get_sitemaps
from ecommercetools.seo.robots import get_robots
from ecommercetools.seo.sitemaps import get_sitemap
from ecommercetools.seo.google_pagespeed_insights import get_core_web_vitals
from ecommercetools.seo.google_knowledge_graph import get_knowledge_graph
from ecommercetools.seo.google_search_console import query_google_search_console
from ecommercetools.seo.google_autocomplete import google_autocomplete
from ecommercetools.seo.google_search import get_indexed_pages
from ecommercetools.seo.google_search import get_serps
from ecommercetools.seo.scraping import scrape_site




import pandas as pd
from transformers import pipeline


def get_summary(text,
                min_length=50,
                max_length=100,
                do_sample=False):
    """Return a summary from a piece of text using a transformer model.

    Args:
        text (string): String of text to summarize. Will be truncated to first 1024 characters.
        min_length (int): Minimum length to return.
        max_length (int): Maximum length to return.
        do_sample (optional, boolean): Set to False to generate unique text or True to extract excerpts.

    Returns:
        string: Summarized text.
    """

    summarizer = pipeline("summarization")
    summary = summarizer(text[:1024],
                         min_length=min_length,
                         max_length=max_length,
                         do_sample=do_sample)
    summary_text = summary[0]['summary_text'].strip().replace(' .', '.')

    return summary_text


def get_summaries(df,
                  text_column,
                  summary_column_name='summary',
                  min_length=50,
                  max_length=100,
                  do_sample=False):
    """Return a summary each of a specified dataframe column using a transformer model.

    Args:
        df (dataframe): Pandas dataframe containing the text to summarize.
        text_column (string): Name of text column to summarize. Will be truncated to first 1024 characters.
        summary_column_name (string, optional): Name of summary column.
        min_length (int, optional): Minimum length to return.
        max_length (int, optional): Maximum length to return.
        do_sample (boolean, optional): Set to False to generate unique text or True to extract excerpts.

    Returns:
        df['summary']: Original dataframe with additional column containing summaries.
    """

    df[summary_column_name] = df.apply(lambda x: get_summary(x[text_column],
                                                             min_length=min_length,
                                                             max_length=max_length,
                                                             do_sample=do_sample), axis=1)
    return df



import re
import random
import itertools
import pandas as pd


def _match_type_exact(keywords):
    exact = []
    for keyword in keywords:
        exact.append([keyword[0], '[' + keyword[1] + ']'])

    df = pd.DataFrame.from_records(exact, columns=['product', 'keywords'])
    df['match_type'] = 'Exact'

    return df


def _match_type_phrase(keywords):
    phrase = []
    for keyword in keywords:
        phrase.append([keyword[0], '"' + keyword[1] + '"'])

    df = pd.DataFrame.from_records(phrase, columns=['product', 'keywords'])
    df['match_type'] = 'Phrase'

    return df


def _match_type_broad(keywords):
    broad = []
    for keyword in keywords:
        broad.append([keyword[0], keyword[1]])

    df = pd.DataFrame.from_records(broad, columns=['product', 'keywords'])
    df['match_type'] = 'Broad'

    return df


def _match_type_broad_modified(keywords):
    broad_modified = []
    for keyword in keywords:
        bmm = ['+' + keyword[1].replace(' ', ' +')]
        broad_modified.append([keyword[0], bmm])

    df = pd.DataFrame.from_records(broad_modified, columns=['product', 'keywords'])
    df['match_type'] = 'Modified'

    return df


def _generate_combinations(products,
                           keywords_prepend,
                           keywords_append):
    """Return a list of all prepended and appended keywords combinations.

    Args:
        products (list): List of product names.
        keywords_prepend (list): List of keywords to prepend to product names.
        keywords_append (list): List of keywords to append to product names.

    Returns:
        keywords (list): List of lists containing the product name and keyword combination.

    Example:
        [['fly rods', 'fly rods'],
        ['fly rods', 'buy fly rods'],
        ['fly rods', 'best fly rods']]
    """

    keywords = []

    for product in products:
        keywords.append([product, product])

        for keyword_prepend in keywords_prepend:
            keywords.append([product, keyword_prepend + ' ' + product])

        for keyword_append in keywords_append:
            keywords.append([product, product + ' ' + keyword_append])

    return keywords


def generate_ad_keywords(products,
                         keywords_prepend,
                         keywords_append,
                         campaign_name):
    """Return a Pandas dataframe of keywords data for use in Google Adwords.

    Args:
        products (list): List of product names.
        keywords_prepend (list): List of keywords to prepend to product names.
        keywords_append (list): List of keywords to append to product names.
        campaign_name (str): Name of paid search campaign.

    Returns:
        df (object): Pandas dataframe containing generated data.
    """

    keywords = _generate_combinations(products, keywords_prepend, keywords_append)

    exact = _match_type_exact(keywords)
    phrase = _match_type_phrase(keywords)
    broad = _match_type_broad(keywords)
    broad_modified = _match_type_broad_modified(keywords)

    df = pd.concat([exact, phrase, broad, broad_modified])
    df['campaign_name'] = campaign_name
    return df


def generate_spintax(text, single=True):
    """Return a list of unique spins of a Spintax text string.

    Args:
        text (string): Spintax text (i.e. I am the {President|King|Ambassador} of Nigeria.)
        single (bool, optional): Optional boolean to return a list or a single spin.

    Returns:
        spins (string, list): Single spin or list of spins depending on single.
    """

    pattern = re.compile('({[^}]+}|[^{}]*)')
    chunks = pattern.split(text)

    def options(s):
        if len(s) > 0 and s[0] == '{':
            return [opt for opt in s[1:-1].split('|')]
        return [s]

    parts_list = [options(chunk) for chunk in chunks]

    spins = []

    for spin in itertools.product(*parts_list):
        spins.append(''.join(spin))

    if single:
        return spins[random.randint(0, len(spins) - 1)]
    else:
        return spins




"""
A very primitive and slow web scraper for SEO tasks on small websites
"""

import requests
import urllib.parse
import pandas as pd
from requests_html import HTMLSession


def _get_source(url: str):
    """Return the source code for the provided URL.

    Args:
        url (string): URL of the page to scrape.

    Returns:
        response (object): HTTP response object from requests_html.
    """

    try:
        session = HTMLSession()
        response = session.get(url)
        return response

    except requests.exceptions.RequestException as e:
        print(e)


def _get_title(response):
    """Parse HTML and extract the title

    :param response: HTML response from Requests-HTML
    :return: HTML element as text
    """

    try:
        return response.html.find('title', first=True).text
    except Exception as e:
        return


def _get_description(response):
    """Parse HTML and extract the title

    :param response: HTML response from Requests-HTML
    :return: HTML element as text
    """

    try:
        return response.html.xpath('//meta[@name="description"]/@content')[0]
    except Exception as e:
        return


def scrape_site(df, url='loc', verbose=False):
    """Scrapes every page in a Pandas dataframe column.

    Args:
        df: Pandas dataframe containing the URL list.
        url (optional, string): Optional name of URL column, if not 'url'
        verbose (optional, boolean, default = False): Set to False to hide progress updates

    Returns:
        df: Pandas dataframe containing all scraped content.

    """

    if verbose:
        pages = len(df)
        minutes = pages / 60

        print('Preparing to scrape ' + str(pages) + ' pages. This will take approximately ' + str(round(minutes)) + ' minutes')

    df_pages = pd.DataFrame(columns=['url', 'title', 'description'])

    for index, row in df.iterrows():

        if verbose:
            print('Scraping: ' + row[url])

        response = _get_source(row[url])

        if response:
            with response as r:
                row = {
                    'url': row[url],
                    'title': _get_title(r),
                    'description': _get_description(r),
                }

                df_pages = df_pages.append(row, ignore_index=True)

    return df_pages




import pandas as pd
import numpy as np
from ecommercetools.utilities import tools


def get_transactions(transaction_items):
    """Return a Pandas DataFrame of transactions from a Pandas DataFrame of transaction items.

    Args:
        transaction_items (object): DataFrame containing order_id, sku, quantity, unit_price, customer_id, order_date

    Returns:
        transactions: Pandas DataFrame containing transactions
    """

    transaction_items = transaction_items.sort_values(by=['order_date'], ascending=True)
    transactions = transaction_items.groupby('order_id').agg(
        order_date=('order_date', 'max'),
        customer_id=('customer_id', 'max'),
        skus=('sku', 'nunique'),
        items=('quantity', 'sum'),
        revenue=('line_price', 'sum'),
    ).reset_index()
    transactions['replacement'] = np.where(transactions['revenue'] > 0, 0, 1)
    transactions['order_number'] = tools.get_cumulative_count(transactions,
                                                              'customer_id',
                                                              'order_id',
                                                              'order_date') + 1
    return transactions





import pandas as pd
import numpy as np
import operator as op
from ecommercetools.transactions import transactions
from ecommercetools import utilities
from sklearn.cluster import KMeans
from sklearn.preprocessing import StandardScaler
from lifetimes import GammaGammaFitter
from lifetimes.utils import summary_data_from_transaction_data
from lifetimes import BetaGeoFitter


def get_customers(transaction_items):
    """Return a Pandas DataFrame of customers from a Pandas DataFrame of transaction items.

    Args:
        transaction_items (object): DataFrame containing order_id, sku, quantity, unit_price, customer_id, order_date

    Returns:
        customers: Pandas DataFrame containing customers
    """

    transactions_df = transactions.get_transactions(transaction_items)
    customers = transactions_df.groupby('customer_id').agg(
        revenue=('revenue', 'sum'),
        orders=('order_id', 'nunique'),
        skus=('skus', 'nunique'),
        items=('items', 'sum'),
        first_order_date=('order_date', 'min'),
        last_order_date=('order_date', 'max')
    ).reset_index()
    customers['avg_items'] = round((customers['items'] / customers['orders']), 2)
    customers['avg_order_value'] = round((customers['revenue'] / customers['orders']), 2)
    customers['tenure'] = (pd.to_datetime('today') - customers['first_order_date']).dt.days
    customers['recency'] = (pd.to_datetime('today') - customers['last_order_date']).dt.days
    customers['cohort'] = customers['first_order_date'].dt.year.astype(str) + \
                          customers['first_order_date'].dt.quarter.astype(str)
    return customers


def _sorted_kmeans(df,
                   metric_column,
                   cluster_name,
                   ascending=True):
    """Runs a K-means clustering algorithm on a specific metric column in a Pandas dataframe.

    Sorts the data in a specified direction; and reassigns cluster numbers to match the data distribution,
    so they are appropriate for RFM segmentation. You may need to log transform heavily skewed data.

    Args:
        df (object): Pandas dataframe
        metric_column (str): Name of metric column
        ascending (bool, optional): Set to False to sort in descending order
        cluster_name (str): Name of cluster

    Returns:
        Original Pandas DataFrame with additional column
    """

    # Fit the model
    kmeans = KMeans(n_clusters=5)
    kmeans.fit(df[[metric_column]])

    # Assign the initial unsorted cluster
    initial_cluster = 'unsorted_' + cluster_name
    df[initial_cluster] = kmeans.predict(df[[metric_column]]) + 1
    df[cluster_name] = df[initial_cluster]

    # Group the clusters and re-rank to determine the correct order
    df_sorted = df.groupby(initial_cluster)[metric_column].mean().round(2).reset_index()
    df_sorted = df_sorted.sort_values(by=metric_column, ascending=ascending).reset_index(drop=True)
    df_sorted[cluster_name] = df_sorted[metric_column].rank(method='max', ascending=ascending).astype(int)

    # Merge data and drop redundant columns
    df = df.merge(df_sorted[[cluster_name, initial_cluster]], on=[initial_cluster])
    df = df.drop(initial_cluster, axis=1)
    df = df.drop(cluster_name + '_x', axis=1)
    df = df.rename(columns={cluster_name + '_y': cluster_name})

    return df


def _label_rfm_segments(rfm):
    """Return a label for a customer based on their RFM score

    Args:
        rfm (int): Full three-digit RFM score, i.e. 555 or 111

    Returns:
        label (str): Descriptive RFM score label, i.e. Risky
    """

    rfm = int(rfm)

    if (rfm >= 111) & (rfm <= 155):
        return 'Risky'

    elif (rfm >= 211) & (rfm <= 255):
        return 'Hold and improve'

    elif (rfm >= 311) & (rfm <= 353):
        return 'Potential loyal'

    elif ((rfm >= 354) & (rfm <= 454)) or ((rfm >= 511) & (rfm <= 535)) or (rfm == 541):
        return 'Loyal'

    elif (rfm == 455) or (rfm >= 542) & (rfm <= 555):
        return 'Star'

    else:
        return 'Other'


def get_rfm_segments(customers):
    """Return a Pandas DataFrame of customer RFM segments from a Pandas DataFrame of customers.

    The DataFrame returned by get_customers() already contains the raw data required, but
    this function will rename it accordingly and use it to assign the customer to a range
    of different segments that can be used for marketing and analysis.

    Args:
        customers: Pandas DataFrame from get_customers()

    Returns:
        segments: Pandas DataFrame

    """

    # Rename the raw data columns
    segments = customers[['customer_id']]
    segments = segments.assign(acquisition_date=customers['first_order_date'])
    segments = segments.assign(recency_date=customers['last_order_date'])
    segments = segments.assign(recency=customers['recency'])
    segments = segments.assign(frequency=customers['orders'])
    segments = segments.assign(monetary=customers['revenue'])
    segments = segments.assign(heterogeneity=customers['skus'])
    segments = segments.assign(tenure=customers['tenure'])

    # Use K-means to create RFMH scores
    segments = _sorted_kmeans(segments, 'recency', 'r', ascending=False)
    segments = _sorted_kmeans(segments, 'frequency', 'f', ascending=True)
    segments = _sorted_kmeans(segments, 'monetary', 'm', ascending=True)
    segments = _sorted_kmeans(segments, 'heterogeneity', 'h', ascending=True)

    # Create scores
    segments = segments.assign(rfm=segments['r'].astype(str) + \
                                   segments['f'].astype(str) + \
                                   segments['m'].astype(str))

    segments = segments.assign(rfm_score=segments['r'].astype(int) + \
                                         segments['f'].astype(int) + \
                                         segments['m'].astype(int))

    # Create labels
    segments['rfm_segment_name'] = segments.apply(lambda x: _label_rfm_segments(x.rfm), axis=1)

    return segments


def _abc_classify_customer(percentage):
    """Apply an ABC classification to each customer based on its ranked percentage revenue contribution.

    Args:
        percentage (float): Cumulative percentage of ranked revenue

    Returns:
        segments: Pandas DataFrame
    """

    if 0 < percentage <= 80:
        return 'A'
    elif 80 < percentage <= 90:
        return 'B'
    else:
        return 'C'


def get_abc_segments(customers,
                     months=12,
                     abc_class_name='abc_class_12m',
                     abc_rank_name='abc_rank_12m'):
    """Return a dataframe containing the ABC class and rank for each customer.

    Apply an ABC classification to each customer based on its ranked percentage revenue contribution.
    This automatically uses a 12 month period by default, but can be modified for other periods to suit.

    Args:

        customers (object): Pandas DataFrame from get_customers()
        months (int, optional): Number of months to use for ABC analysis (12 by default)
        abc_class_name (str, optional): Name to assign to ABC class string (abc_class_12m by default)
        abc_rank_name (str, optional): Name to assign to ABC rank string (abc_rank_12m by default)

    Returns:
        abc: Pandas DataFrame
    """

    # Calculate data for customers who purchased within the specified period
    purchased = customers[customers['recency'] <= (months * 30)]
    purchased = purchased.sort_values(by='revenue', ascending=False)
    purchased['revenue_cumsum'] = purchased['revenue'].cumsum()
    purchased['revenue_total'] = purchased['revenue'].sum()
    purchased['revenue_running_percentage'] = (purchased['revenue_cumsum'] / purchased['revenue_total']) * 100
    purchased[abc_class_name] = purchased['revenue_running_percentage'].apply(_abc_classify_customer)
    purchased[abc_rank_name] = purchased['revenue_running_percentage'].rank().astype(int)
    purchased.drop(['revenue_cumsum', 'revenue_total', 'revenue_running_percentage'], axis=1, inplace=True)

    # Assign lapsed customers to class D
    lapsed = customers[customers['recency'] > (months * 30)]

    # Return ABC segments
    abc = purchased.append(lapsed)
    abc[abc_class_name].fillna('D', inplace=True)
    abc[abc_rank_name].fillna(len(purchased) + 1, inplace=True)
    abc = abc[['customer_id', abc_class_name, abc_rank_name]]
    return abc


def get_cohorts(df, period='M'):
    """Return a customer cohort matrix from a dataframe of transactional items.

    Given a Pandas DataFrame of transactional items, this function returns
    a Pandas DataFrame containing the acquisition cohort and order cohort which
    can be used for customer analysis or the creation of a cohort analysis matrix.

    Args:
        df (object): Pandas DataFrame. Required columns: order_id, customer_id, order_date.
        period (str, optional): Period value - M, Q, or Y. Create cohorts using month, quarter, or year of acquisition.

    Returns:
        df (object): Pandas DataFrame
    """

    df = df[['customer_id', 'order_id', 'order_date']].drop_duplicates()
    df = df.assign(acquisition_cohort=df.groupby('customer_id') \
        ['order_date'].transform('min').dt.to_period(period))
    df = df.assign(order_cohort=df['order_date'].dt.to_period(period))
    return df


def get_retention(df, period='M'):
    """Calculate the retention of customers in each month after their acquisition.

    Args:
        df (object): Pandas DataFrame. Required columns: order_id, customer_id, order_date.
        period (str, optional): Period value - M, Q, or Y. Create cohorts using month, quarter, or year of acquisition.

    Returns:
    -------
        df (object): Pandas DataFrame
    """

    df = get_cohorts(df, period).groupby(['acquisition_cohort', 'order_cohort']) \
        .agg(customers=('customer_id', 'nunique')) \
        .reset_index(drop=False)
    df['periods'] = (df.order_cohort - df.acquisition_cohort) \
        .apply(op.attrgetter('n'))

    return df


def get_cohort_matrix(df, period='M', percentage=False):
    """Return a cohort matrix showing the number of customers who purchased in each period after their acquisition.

    Args:
        df (object): Pandas DataFrame. Required columns: order_id, customer_id, order_date.
        period (str, optional): Period value - M, Q, or Y. Create cohorts using month, quarter, or year of acquisition.
        percentage (bool, optional): True or False. Return raw numbers or a percentage retention.

    Returns:
        df (object): Pandas DataFrame
    """

    df = get_retention(df, period).pivot_table(index='acquisition_cohort',
                                               columns='periods',
                                               values='customers')

    if percentage:
        df = df.divide(df.iloc[:, 0], axis=0)

    return df


def _days_to_next_order(avg_latency, std_latency, recency):
    """Estimate the number of days to a customer's next order using latency.

    Args:
        avg_latency (float): Average latency in days
        std_latency (float): Standard deviation of latency in days
        recency (float): Recency in days
    Returns:
        Approximate number of days until the next order.
    """

    return avg_latency - (recency - std_latency)


def _latency_label_customers(avg_latency, std_latency, recency):
    """Add a label to describe a customer's latency metric.

    Args:
        avg_latency (float): Average latency in days
        std_latency (float): Standard deviation of latency in days
        recency (float): Recency in days
    Returns:
           Label describing the latency metric in relation to the customer.
    """

    days_to_next_order_upper = avg_latency - (recency - std_latency)
    days_to_next_order_lower = avg_latency - (recency + std_latency)

    if recency < days_to_next_order_lower:
        return 'Order not due'

    elif (recency <= days_to_next_order_lower) or (recency <= days_to_next_order_upper):
        return 'Order due soon'

    elif recency > days_to_next_order_upper:
        return 'Order overdue'

    else:
        return 'Not sure'


def get_latency(df_transactions):
    """Return a Pandas dataframe containing latency metrics for each customer.

    Args:
        df_transactions: Pandas dataframe from get_transactions().

    Returns:
        Pandas dataframe of customer purchase latency metrics.
    """

    # Create latency dataframe and calculate granular metrics
    df_latency = df_transactions[['order_id', 'customer_id', 'order_date', 'revenue']]
    df_latency = df_latency[df_latency['revenue'] > 0]
    df_latency = df_latency.sort_values(by=['order_date'], ascending=False)
    df_latency['prev_order_date'] = utilities.get_previous_value(df_latency, 'customer_id', 'order_date')
    df_latency['days_since_prev_order'] = utilities.get_days_since_date(df_latency, 'prev_order_date', 'order_date')
    df_latency['order_number'] = utilities.get_cumulative_count(df_latency, 'customer_id', 'order_id', 'order_date')

    # Create customer dataframe and calculate aggregate metrics
    df_customers = pd.DataFrame(df_latency['customer_id'].unique())
    df_customers.columns = ['customer_id']

    # Calculate frequency
    df_frequency = df_latency.groupby('customer_id')['order_id'].nunique().reset_index()
    df_frequency.columns = ['customer_id', 'frequency']
    df_customers = df_customers.merge(df_frequency, on='customer_id')

    # Calculate recency
    df_recency = df_latency.groupby('customer_id')['order_date'].max().reset_index()
    df_recency.columns = ['customer_id', 'recency_date']
    df_customers = df_customers.merge(df_recency, on='customer_id')
    df_customers['recency'] = round((pd.to_datetime('today') - df_customers['recency_date']) \
                                    / np.timedelta64(1, 'D')).astype(int)

    # Calculate average latency
    df_avg_latency = df_latency.groupby('customer_id')['days_since_prev_order'].mean().astype(int).reset_index()
    df_avg_latency.columns = ['customer_id', 'avg_latency']
    df_customers = df_customers.merge(df_avg_latency, on='customer_id')

    # Calculate standard deviation of latency for returning customers
    df_latency_returning = df_latency[df_latency['order_number'] > 0]

    # Min latency
    df_min = df_latency_returning.groupby('customer_id')['days_since_prev_order'].min().astype(int).reset_index()
    df_min.columns = ['customer_id', 'min_latency']
    df_customers = df_customers.merge(df_min, on='customer_id')

    # Max latency
    df_max = df_latency_returning.groupby('customer_id')['days_since_prev_order'].max().astype(int).reset_index()
    df_max.columns = ['customer_id', 'max_latency']
    df_customers = df_customers.merge(df_max, on='customer_id')

    # STD latency
    df_std = df_latency_returning.groupby('customer_id')['days_since_prev_order'].std().reset_index()
    df_std.columns = ['customer_id', 'std_latency']
    df_customers = df_customers.merge(df_std, on='customer_id')

    # Coefficient of Variation of latency
    df_customers['cv'] = df_customers['std_latency'] / df_customers['avg_latency']

    # Calculate approximate days to next order
    df_customers['days_to_next_order'] = df_customers.apply(
        lambda x: _days_to_next_order(x['avg_latency'], x['std_latency'], x['recency']), axis=1).round()

    # Label latency
    df_customers['label'] = df_customers.apply(
        lambda x: _latency_label_customers(x['avg_latency'], x['std_latency'], x['recency']), axis=1)

    return df_customers


def _get_lifetimes_rfmt(df_transactions, observation_period_end):
    """Return the RFMT data from the Lifetimes model.

    Args:
        df_transactions (df): Pandas dataframe of transactions from get_transactions()
        observation_period_end (string): Date string in YYYY-MM-DD format representing end of observation period.

    Returns:
        df: Pandas dataframe containing frequency, recency, T, monetary_value per customer.
    """

    df_transactions = df_transactions[df_transactions['replacement'] == 0]

    df = summary_data_from_transaction_data(df_transactions,
                                            'customer_id',
                                            'order_date',
                                            'revenue',
                                            observation_period_end=observation_period_end)
    return df


def _get_predicted_purchases(df_transactions,
                             observation_period_end,
                             days=90):
    """Return the number of predicted purchases per customer from the Lifetimes BG/NBD model.

    Args:
        df_transactions (df): Pandas dataframe of transactions from get_transactions()
        observation_period_end (string): Date string in YYYY-MM-DD format representing end of observation period.

    Returns:
        df: Pandas dataframe containing frequency, recency, T, monetary_value per customer, and predicted purchases.
    """

    df = _get_lifetimes_rfmt(df_transactions, observation_period_end)
    bgf = BetaGeoFitter(penalizer_coef=0)
    bgf.fit(df['frequency'], df['recency'], df['T'])
    df['predicted_purchases'] = bgf.conditional_expected_number_of_purchases_up_to_time(days,
                                                                                        df['frequency'],
                                                                                        df['recency'],
                                                                                        df['T'])
    return df


def _get_predicted_aov(df_transactions,
                       observation_period_end,
                       ggf_penalizer_coef=0):
    """Returns the predicted AOV for each customer via the Gamma-Gamma model.
    This function uses models from the Lifetimes package.

    Args:
        df_transactions (df): Pandas dataframe of transactions from get_transactions()
        observation_period_end (string): Date string in YYYY-MM-DD format for end of observation period.
        ggf_penalizer_coef (float, optional): Penalizer coefficient for Gamma-Gamma model. See Lifetimes.

    Returns:
        Predicted AOV for each customer.
    """

    df_rfmt = _get_lifetimes_rfmt(df_transactions, observation_period_end)

    df_returning = df_rfmt[df_rfmt['frequency'] > 0]
    df_returning = df_rfmt[df_rfmt['monetary_value'] > 0]

    ggf = GammaGammaFitter(penalizer_coef=ggf_penalizer_coef)
    ggf.fit(df_returning['frequency'],
            df_returning['monetary_value'])

    predicted_monetary = ggf.conditional_expected_average_profit(
        df_returning['frequency'],
        df_returning['monetary_value']
    )

    aov_df = pd.DataFrame(predicted_monetary, columns=['aov'])

    return aov_df


def _get_predicted_clv(df_transactions,
                       observation_period_end,
                       months=12,
                       discount_rate=0.01,
                       ggf_penalizer_coef=0,
                       bgf_penalizer_coef=0):
    """Return the predicted CLV for each customer using the Gamma-Gamma and BG/NBD models.
    This function uses models from the Lifetimes package.

    Args:
        df_transactions (df): Pandas dataframe of transactions from get_transactions()
        observation_period_end (string): Date string in YYYY-MM-DD format for end of observation period.
        months (int, optional): Optional number of months in CLV prediction window.
        discount_rate (float, optional): Discount rate. See Lifetimes.
        ggf_penalizer_coef (float, optional): Penalizer coefficient for Gamma-Gamma model. See Lifetimes.
        bgf_penalizer_coef (float, optional): Penalizer coefficient for BG/NBD model. See Lifetimes.

    Returns:
        Predicted CLV for each customer.
    """

    df_rfmt = _get_lifetimes_rfmt(df_transactions, observation_period_end)
    df_returning = df_rfmt[df_rfmt['frequency'] > 0]
    df_returning = df_rfmt[df_rfmt['monetary_value'] > 0]

    ggf = GammaGammaFitter(penalizer_coef=ggf_penalizer_coef)
    ggf.fit(df_returning['frequency'],
            df_returning['monetary_value'])

    bgf = BetaGeoFitter(penalizer_coef=bgf_penalizer_coef)
    bgf.fit(df_returning['frequency'],
            df_returning['recency'],
            df_returning['T'])

    preds = ggf.customer_lifetime_value(
        bgf,
        df_returning['frequency'],
        df_returning['recency'],
        df_returning['T'],
        df_returning['monetary_value'],
        time=months,
        discount_rate=discount_rate
    ).to_frame().reset_index()

    return preds


def get_customer_predictions(df_transactions,
                             observation_period_end,
                             days=90,
                             months=3,
                             discount_rate=0.01,
                             ggf_penalizer_coef=0,
                             bgf_penalizer_coef=0):
    """Get predicted customer purchased, AOV, and CLV for the defined period.

    This uses the Lifetimes package to run the Gamma-Gamma, and BG/NBD models
    and predict the AOV, CLV, and number of purchases each customer will make.
    These models use a different approach to measuring RFMT than the other
    functions in EcommerceTools, so are not directly comparable, so the results
    have been removed from the output.

    Args:
        df_transactions (df): Pandas dataframe of transactions from get_transactions()
        observation_period_end (string): Date string in YYYY-MM-DD format for end of observation period.
        days (int, optional): Optional number of days in purchase prediction window.
        months (int, optional): Optional number of months in CLV prediction window.
        discount_rate (float, optional): Discount rate. See Lifetimes.
        ggf_penalizer_coef (float, optional): Penalizer coefficient for Gamma-Gamma model. See Lifetimes.
        bgf_penalizer_coef (float, optional): Penalizer coefficient for BG/NBD model. See Lifetimes.

    Returns:
        df_predictions: Pandas dataframe containing predictions from Gamma-Gamma and BG/NBD models.
    """

    df_predicted_purchases = _get_predicted_purchases(df_transactions,
                                                      observation_period_end,
                                                      days=days)
    df_aov = _get_predicted_aov(df_transactions,
                                observation_period_end)

    df_clv = _get_predicted_clv(df_transactions,
                                observation_period_end,
                                months=months,
                                discount_rate=discount_rate,
                                bgf_penalizer_coef=bgf_penalizer_coef,
                                ggf_penalizer_coef=ggf_penalizer_coef
                                )

    df_predictions = df_predicted_purchases.merge(df_aov, on='customer_id', how='left')
    df_predictions = df_predictions.merge(df_clv, on='customer_id', how='left')

    return df_predictions[['customer_id', 'predicted_purchases', 'aov', 'clv']]

